@@ -20,7 +20,6 @@ let annot_schema = Ocaml.annot_schema_of_target target
 type param = {
   deref : (Ocaml.Repr.t, Json.json_repr) Mapping.mapping ->
     (Ocaml.Repr.t, Json.json_repr) Mapping.mapping;
-  std : bool;
   unknown_field_handler : string option;
   (* Optional handler that takes a field name as argument
      and does something with it such as displaying a warning message. *)
@@ -34,40 +33,35 @@ type param = {
 
 }
 
-let make_ocaml_json_intf ~with_create buf deref defs =
-  List.concat_map snd defs
-  |> List.filter Ox_emit.include_intf
-  |> List.iter (fun x ->
-    let s = x.def_name in
-    let full_name = Ox_emit.get_full_type_name x in
-    let writer_params =
-      String.concat "" (
-        List.map
-          (fun s -> sprintf "\n  (Buffer.t -> '%s -> unit) ->" s)
-          x.def_param
-      )
-    in
-    let reader_params =
-      String.concat "" (
-        List.map
-          (fun s ->
-             sprintf "\n  (Yojson.Safe.lexer_state -> \
-                      Lexing.lexbuf -> '%s) ->" s)
-          x.def_param
-      )
-    in
+let writer_parameters params =
+  String.concat "" (
+    List.map
+      (fun s -> sprintf "\n  (Buffer.t -> '%s -> unit) ->" s)
+      params
+  )
+let reader_parameters params =
+  String.concat "" (
+    List.map
+      (fun s ->
+         sprintf "\n  (Yojson.Safe.lexer_state -> \
+                  Lexing.lexbuf -> '%s) ->" s)
+      params
+  )
+
+let val_write buf fun_name writer_params full_name short_name =
     bprintf buf "\
-val write_%s :%s
+val %s :%s
   Buffer.t -> %s -> unit
   (** Output a JSON value of type {!type:%s}. *)
 
 "
-      s writer_params
+      fun_name writer_params
       full_name
-      s;
+      short_name
 
+let val_string_of buf fun_name writer_params full_name short_name =
     bprintf buf "\
-val string_of_%s :%s
+val %s :%s
   ?len:int -> %s -> string
   (** Serialize a value of type {!type:%s}
       into a JSON string.
@@ -76,29 +70,48 @@ val string_of_%s :%s
                  Default: 1024. *)
 
 "
-      s writer_params
+      fun_name writer_params
       full_name
-      s;
+      short_name
 
+let val_read buf fun_name reader_params full_name short_name =
     bprintf buf "\
-val read_%s :%s
+val %s :%s
   Yojson.Safe.lexer_state -> Lexing.lexbuf -> %s
   (** Input JSON data of type {!type:%s}. *)
 
 "
-      s reader_params
+      fun_name reader_params
       full_name
-      s;
+      short_name
 
+let val_of_string buf fun_name reader_params full_name short_name =
     bprintf buf "\
-val %s_of_string :%s
+val %s :%s
   string -> %s
   (** Deserialize JSON data of type {!type:%s}. *)
 
 "
-      s reader_params
+      fun_name reader_params
       full_name
-      s;
+      short_name
+
+let make_ocaml_json_intf ~with_create buf deref defs =
+  List.concat_map snd defs
+  |> List.filter Ox_emit.include_intf
+  |> List.iter (fun x ->
+    let s = x.def_name in
+    let full_name = Ox_emit.get_full_type_name x in
+    let writer_params = writer_parameters x.def_param in
+    let reader_params = reader_parameters x.def_param in
+    val_write buf (sprintf "write_%s" s)
+      writer_params full_name s;
+    val_string_of buf (sprintf "string_of_%s" s)
+      writer_params full_name s;
+    val_read buf (sprintf "read_%s" s)
+      reader_params full_name s;
+    val_of_string buf (sprintf "%s_of_string" s)
+      reader_params full_name s;
     Ox_emit.maybe_write_creator_intf ~with_create deref buf x
   )
 
@@ -143,14 +156,9 @@ let rec get_writer_name
 
   | Float (_, Float, Float j) ->
       (match j with
-         Float None ->
-           if p.std then "Yojson.Safe.write_std_float"
-           else "Yojson.Safe.write_float"
+         Float None -> "Yojson.Safe.write_std_float"
        | Float (Some precision) ->
-           if p.std then
-             sprintf "Yojson.Safe.write_std_float_prec %i" precision
-           else
-             sprintf "Yojson.Safe.write_float_prec %i" precision
+           sprintf "Yojson.Safe.write_std_float_prec %i" precision
        | Int ->
            "Atdgen_runtime.Oj_run.write_float_as_int"
       )
@@ -334,10 +342,7 @@ let rec make_writer ?type_constraint p (x : Oj_mapping.t) : Indent.t list =
       let l =
         insert (Line "Buffer.add_char ob ',';") (Array.to_list a)
       in
-      let op, cl =
-        if p.std then '[', ']'
-        else '(', ')'
-      in
+      let op, cl = '[', ']' in
       [
         Annot ("fun", Line "fun ob x ->");
         Block [
@@ -379,8 +384,7 @@ let rec make_writer ?type_constraint p (x : Oj_mapping.t) : Indent.t list =
 
   | Option (_, x, Option, Option) ->
       [
-        Line (sprintf "Atdgen_runtime.Oj_run.write_%soption ("
-                (if p.std then "std_" else ""));
+        Line "Atdgen_runtime.Oj_run.write_std_option (";
         Block (make_writer p x);
         Line ")";
       ]
@@ -420,14 +424,10 @@ and make_variant_writer p ~tick ~open_enum x : Indent.t list =
   let json_cons = j.Json.json_cons in
   match x.var_arg with
   | None ->
-      let enclose s =
-        if p.std then s
-        else "<" ^ s ^ ">"
-      in
       [
         Line (sprintf "| %s%s -> Buffer.add_string ob %S"
                 tick ocaml_cons
-                (enclose (make_json_string json_cons)))
+                (make_json_string json_cons))
       ]
   | Some v when open_enum ->
       (* v should resolve to type string. *)
@@ -440,10 +440,7 @@ and make_variant_writer p ~tick ~open_enum x : Indent.t list =
       ]
   | Some (Record (_, a, Record o, Record _)) ->
       (* We need a special case because inline-records cannot escape their scope. *)
-      let op, sep, cl =
-        if p.std then "[", ",", ']'
-        else "<", ":", '>'
-      in
+      let op, sep, cl = "[", ",", ']' in
       [
         Line (sprintf "| %s%s x ->" tick ocaml_cons);
         Block [
@@ -456,10 +453,7 @@ and make_variant_writer p ~tick ~open_enum x : Indent.t list =
         ]
       ]
   | Some v ->
-      let op, sep, cl =
-        if p.std then "[", ",", ']'
-        else "<", ":", '>'
-      in
+      let op, sep, cl = "[", ",", ']' in
       [
         Line (sprintf "| %s%s x ->" tick ocaml_cons);
         Block [
@@ -649,10 +643,6 @@ let rec make_reader p type_annot (x : Oj_mapping.t) : Indent.t list =
       let fallback_expr =
         [ Line "Atdgen_runtime.Oj_run.invalid_variant_tag p x" ]
       in
-      let cases =
-        make_cases_reader p type_annot
-          ~tick ~open_enum ~std:false ~fallback_expr l
-      in
       let l0, l1 =
         List.partition (fun x -> x.var_arg = None || open_enum) l
       in
@@ -667,16 +657,8 @@ let rec make_reader p type_annot (x : Oj_mapping.t) : Indent.t list =
       let read_tag =
         [
           Line "Yojson.Safe.read_space p lb;";
-          Line "match Yojson.Safe.start_any_variant p lb with";
+          Line "match Atdgen_runtime.Yojson_extra.start_any_variant p lb with";
           Block [
-            Line "| `Edgy_bracket -> (";
-            Block [
-              Block [
-                Line "match Yojson.Safe.read_ident p lb with";
-                Block cases;
-              ];
-              Line ")";
-            ];
             Line "| `Double_quote -> (";
             Block [
               Block [
@@ -1094,7 +1076,7 @@ and make_tuple_reader p a =
                   Line "in";
                   Line "incr len;";
                   Line "Yojson.Safe.read_space p lb;";
-                  Line "Yojson.Safe.read_tuple_sep2 p std_tuple lb;";
+                  Line "Atdgen_runtime.Yojson_extra.read_tuple_sep2 p lb;";
                   Line "x"
                 ]
               else if i = min_length - 1 then
@@ -1106,9 +1088,9 @@ and make_tuple_reader p a =
                   Line "(try";
                   Block [
                     Line "Yojson.Safe.read_space p lb;";
-                    Line "Yojson.Safe.read_tuple_sep2 p std_tuple lb;";
+                    Line "Atdgen_runtime.Yojson_extra.read_tuple_sep2 p lb;";
                   ];
-                  Line "with Yojson.End_of_tuple -> end_of_tuple := true);";
+                  Line "with Atdgen_runtime.Yojson_extra.End_of_tuple -> end_of_tuple := true);";
                   Line "x"
                 ]
               else
@@ -1124,9 +1106,9 @@ and make_tuple_reader p a =
                     Line "(try";
                     Block [
                       Line "Yojson.Safe.read_space p lb;";
-                      Line "Yojson.Safe.read_tuple_sep2 p std_tuple lb;";
+                      Line "Atdgen_runtime.Yojson_extra.read_tuple_sep2 p lb;";
                     ];
-                    Line "with Yojson.End_of_tuple ->";
+                    Line "with Atdgen_runtime.Yojson_extra.End_of_tuple ->";
                     Block [
                       Line "end_of_tuple := true);";
                     ];
@@ -1163,8 +1145,8 @@ and make_tuple_reader p a =
   let finish_empty_tuple =
     if min_length = 0 then
       [
-        Line "(try Yojson.Safe.read_tuple_end2 p std_tuple lb";
-        Line "with Yojson.End_of_tuple -> end_of_tuple := true)";
+        Line "(try Atdgen_runtime.Yojson_extra.read_tuple_end2 p lb";
+        Line "with Atdgen_runtime.Yojson_extra.End_of_tuple -> end_of_tuple := true)";
       ]
     else []
   in
@@ -1179,11 +1161,11 @@ and make_tuple_reader p a =
           Block [
             Line "Yojson.Safe.skip_json p lb;";
             Line "Yojson.Safe.read_space p lb;";
-            Line "Yojson.Safe.read_tuple_sep2 p std_tuple lb;";
+            Line "Atdgen_runtime.Yojson_extra.read_tuple_sep2 p lb;";
           ];
           Line "done";
         ];
-        Line "with Yojson.End_of_tuple -> ()";
+        Line "with Atdgen_runtime.Yojson_extra.End_of_tuple -> ()";
       ];
       Line ");"
     ]
@@ -1191,7 +1173,7 @@ and make_tuple_reader p a =
 
   [
     Line "Yojson.Safe.read_space p lb;";
-    Line "let std_tuple = Yojson.Safe.start_any_tuple p lb in";
+    Line "Atdgen_runtime.Yojson_extra.start_any_tuple p lb;";
     Line "let len = ref 0 in";
     Line "let end_of_tuple = ref false in";
     Inline finish_empty_tuple;
@@ -1201,7 +1183,7 @@ and make_tuple_reader p a =
       Inline skip_remaining_cells;
       Line make_tuple;
     ];
-    Line "with Yojson.End_of_tuple ->";
+    Line "with Atdgen_runtime.Yojson_extra.End_of_tuple ->";
     Block [
       Line (sprintf
               "Atdgen_runtime.Oj_run.missing_tuple_fields p !len %s);"
@@ -1281,13 +1263,12 @@ let make_ocaml_json_reader p ~original_types is_rec let1 let2 def =
 
 
 let make_ocaml_json_impl
-    ~std ~unknown_field_handler
+    ~unknown_field_handler
     ~with_create ~force_defaults ~preprocess_input ~original_types
     ~ocaml_version
     buf deref defs =
   let p =
     { deref
-    ; std
     ; unknown_field_handler
     ; force_defaults
     ; preprocess_input
@@ -1332,7 +1313,7 @@ let make_mli
 
 let make_ml
     ~header ~opens ~with_typedefs ~with_create ~with_fundefs
-    ~std ~unknown_field_handler
+    ~unknown_field_handler
     ~force_defaults ~preprocess_input ~original_types
     ~ocaml_version
     ocaml_typedefs deref defs =
@@ -1345,19 +1326,62 @@ let make_ml
     bprintf buf "\n";
   if with_fundefs then
     make_ocaml_json_impl
-      ~std ~unknown_field_handler
+      ~unknown_field_handler
       ~with_create ~force_defaults ~preprocess_input ~original_types
       ~ocaml_version
       buf deref defs;
   Buffer.contents buf
 
+let make_extra_generic_modules definitions =
+  let ml = Buffer.create 42 in
+  let mli = Buffer.create 42 in
+  let oof buf fmt = Format.kasprintf (Buffer.add_string buf) fmt in
+  let mlf fmt = oof ml fmt in
+  let mlif fmt = oof mli fmt in
+  let bothf fmt=
+    Format.kasprintf (fun s -> 
+      Buffer.add_string ml s; Buffer.add_string mli s) fmt in
+  bothf "\n\n(** {3 Generic Modules } *)\n";
+  let make_module name params =
+    bothf "module %s" (String.capitalize_ascii name);
+    mlf " = struct\n";
+    mlif " : sig\n";
+    let type_params =
+      List.map (sprintf "'%s") params |> String.concat ", " in
+    let type_params_prefix =
+      match params with [] -> "" | _ -> sprintf "(%s) " type_params in
+    bothf "type nonrec %st = %s%s\n" type_params_prefix type_params_prefix name;
+    mlf "let write = write_%s\n" name;
+    mlf "let read = read_%s\n" name;
+    mlf "let to_string = string_of_%s\n" name;
+    mlf "let of_string = %s_of_string\n" name;
+    val_write mli (sprintf "write" ) (writer_parameters params) 
+      (type_params_prefix ^ name) name;
+    val_string_of mli (sprintf "to_string" ) (writer_parameters params) 
+      (type_params_prefix ^ name) name;
+    val_read mli (sprintf "read" ) (reader_parameters params) 
+      (type_params_prefix ^ name) name;
+    val_of_string mli (sprintf "of_string" ) (reader_parameters params) 
+      (type_params_prefix ^ name) name;
+    bothf "end\n"
+  in
+  List.iter
+    (fun (_, body) ->
+      List.iter
+        (fun (Atd.Ast.Type (_, (name, params, _), type_expr)) ->
+          match type_expr with
+          | Name (_, (_, "abstract", _), _) -> ()
+          | _ -> make_module name params)
+        body)
+    definitions;
+  (`Ml (Buffer.contents ml), `Mli (Buffer.contents mli))
+
 let make_ocaml_files
     ~opens
     ~with_typedefs
     ~with_create
     ~with_fundefs
     ~all_rec
-    ~std
     ~unknown_field_handler
     ~pos_fname
     ~pos_lnum
@@ -1366,6 +1390,7 @@ let make_ocaml_files
     ~preprocess_input
     ~ocaml_version
     ~pp_convs
+    ~add_generic_modules
     atd_file out =
   let ((head, m0), _) =
     match atd_file with
@@ -1417,9 +1442,15 @@ let make_ocaml_files
   in
   let ml =
     make_ml ~header ~opens ~with_typedefs ~with_create ~with_fundefs
-      ~std ~unknown_field_handler
+      ~unknown_field_handler
       ~force_defaults ~preprocess_input ~original_types
       ~ocaml_version
       ocaml_typedefs (Mapping.make_deref defs) defs
   in
+  let mli, ml =
+    if not add_generic_modules then mli, ml
+    else
+      let `Ml extra_ml, `Mli extra_mli = make_extra_generic_modules m1 in
+      (mli ^ extra_mli, ml ^ extra_ml)
+  in
   Ox_emit.write_ocaml out mli ml
