@@ -192,13 +192,42 @@ let single_esc s =
 let _double_esc s =
   escape_string_content Double s
 
-let fixed_size_preamble atd_filename =
-  sprintf {|"""Generated by atdpy from type definitions in %s.
+let not_implemented loc msg =
+  A.error_at loc ("not implemented in atdpy: " ^ msg)
 
-This implements classes for the types defined in '%s', providing
-methods and functions to convert data from/to JSON.
-"""
+let todo hint =
+  failwith ("TODO: " ^ hint)
 
+let spaced ?(spacer = [Line ""]) (blocks : B.node list) : B.node list =
+  let rec spaced xs =
+    match List.filter (fun x -> not (B.is_empty_node x)) xs with
+    | []
+    | [_] as xs -> xs
+    | a :: rest -> a :: spacer @ spaced rest
+  in
+  spaced blocks
+
+let double_spaced blocks =
+  spaced ~spacer:[Line ""; Line ""] blocks
+
+(** [blank_line_before nodes] inserts [n] blank lines iff [nodes]
+      is empty. *)
+let blank_lines_before ?(n = 1) nodes =
+  if is_empty nodes then []
+  else List.init n (fun _ -> Line "")
+
+let preamble ~atd_filename ~user_docstring = [
+  Line (sprintf {|"""Generated by atdpy from type definitions in %s.|}
+          atd_filename);
+  Line "";
+  Line (sprintf
+          "This implements classes for the types defined in '%s', providing
+methods and functions to convert data from/to JSON."
+          atd_filename);
+  Inline (blank_lines_before user_docstring);
+  Inline user_docstring;
+  Line {|"""|};
+  Line {|
 # Disable flake8 entirely on this file:
 # flake8: noqa
 
@@ -453,26 +482,7 @@ def _atd_write_option(write_elt: Callable[[Any], Any]) \
 ############################################################################
 # Public classes
 ############################################################################|}
-    atd_filename
-    atd_filename
-
-let not_implemented loc msg =
-  A.error_at loc ("not implemented in atdpy: " ^ msg)
-
-let todo hint =
-  failwith ("TODO: " ^ hint)
-
-let spaced ?(spacer = [Line ""]) (blocks : B.node list) : B.node list =
-  let rec spaced xs =
-    match List.filter (fun x -> not (B.is_empty_node x)) xs with
-    | []
-    | [_] as xs -> xs
-    | a :: rest -> a :: spacer @ spaced rest
-  in
-  spaced blocks
-
-let double_spaced blocks =
-  spaced ~spacer:[Line ""; Line ""] blocks
+]
 
 (*
    Representations of ATD type '(string * value) list' in JSON and Python.
@@ -513,34 +523,42 @@ let py_type_name env (name : string) =
   | "abstract" -> "Any"
   | user_defined -> class_name env user_defined
 
-let rec type_name_of_expr env (e : type_expr) : string =
+(* Return the mypy type name for type expression that doesn't require
+   us to create a new class (type aliases).
+
+   Also return the outermost annotation.
+*)
+let rec type_name_of_expr env (e : type_expr) : string * (loc * annot) =
   match e with
   | Sum (loc, _, _) -> not_implemented loc "inline sum types"
   | Record (loc, _, _) -> not_implemented loc "inline records"
   | Tuple (loc, xs, an) ->
       let type_names =
         xs
-        |> List.map (fun (loc, x, an) -> type_name_of_expr env x)
+        |> List.map (fun (loc, x, an) -> type_name_of_expr env x |> fst)
       in
-      sprintf "Tuple[%s]" (String.concat ", " type_names)
+      sprintf "Tuple[%s]" (String.concat ", " type_names), (loc, an)
   | List (loc, e, an) ->
      (match assoc_kind loc e an with
        | Array_list
        | Object_list _ ->
            sprintf "List[%s]"
-             (type_name_of_expr env e)
+             (type_name_of_expr env e |> fst)
        | Array_dict (key, value) ->
            sprintf "Dict[%s, %s]"
-             (type_name_of_expr env key) (type_name_of_expr env value)
+             (type_name_of_expr env key |> fst)
+             (type_name_of_expr env value |> fst)
        | Object_dict value ->
            sprintf "Dict[str, %s]"
-             (type_name_of_expr env value)
-      )
-  | Option (loc, e, an) -> sprintf "Optional[%s]" (type_name_of_expr env e)
-  | Nullable (loc, e, an) -> sprintf "Optional[%s]" (type_name_of_expr env e)
+             (type_name_of_expr env value |> fst)
+      ), (loc, an)
+  | Option (loc, e, an) ->
+      sprintf "Optional[%s]" (type_name_of_expr env e |> fst), (loc, an)
+  | Nullable (loc, e, an) ->
+      sprintf "Optional[%s]" (type_name_of_expr env e |> fst), (loc, an)
   | Shared (loc, e, an) -> not_implemented loc "shared"
   | Wrap (loc, e, an) -> todo "wrap"
-  | Name (loc, (loc2, name, []), an) -> py_type_name env name
+  | Name (loc, (loc2, name, []), an) -> py_type_name env name, (loc, an)
   | Name (loc, (_, name, _::_), _) -> assert false
   | Tvar (loc, _) -> not_implemented loc "type variables"
 
@@ -683,6 +701,65 @@ let construct_json_field env trans_meth
         Block assignment
       ]
 
+let class_doc_paragraph_width = 74
+
+let trans_field_doc_to_docstring
+    env trans_meth ((loc, (name, kind, an), e) : simple_field) : node list =
+  match Atd.Doc.get_doc loc an with
+  | None -> []
+  | Some blocks ->
+      (* Insert a prefix containing the field name (Sphinx format) *)
+      let prefix =
+        sprintf ":param %s: " (inst_var_name trans_meth name)
+      in
+      let blocks =
+         match blocks with
+          | Paragraph elements :: blocks ->
+              Atd.Doc.Paragraph (Text prefix :: elements) :: blocks
+          | _ -> Paragraph [Text prefix] :: blocks
+      in
+      Codegen_doc.make_unquoted_multiline_docstring
+        ~paragraph_width:class_doc_paragraph_width
+        blocks
+
+let trans_any_doc_to_docstring ~paragraph_width loc an : node list =
+  match Atd.Doc.get_doc loc an with
+  | None -> []
+  | Some blocks ->
+      Codegen_doc.make_unquoted_multiline_docstring
+        ~paragraph_width:class_doc_paragraph_width
+        blocks
+
+let trans_case_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:class_doc_paragraph_width
+    loc an
+
+let trans_record_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:class_doc_paragraph_width
+    loc an
+
+let trans_sum_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:class_doc_paragraph_width
+     loc an
+
+let trans_alias_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:class_doc_paragraph_width
+     loc an
+
+let trans_def_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:class_doc_paragraph_width
+     loc an
+
+let trans_head_doc_to_docstring loc an =
+  trans_any_doc_to_docstring
+    ~paragraph_width:78
+    loc an
+
 (*
    Function value that can be applied to a JSON node, converting it
    to the desired value.
@@ -784,7 +861,7 @@ let from_json_class_argument
 let inst_var_declaration
     env trans_meth ((loc, (name, kind, an), e) : simple_field) =
   let var_name = inst_var_name trans_meth name in
-  let type_name = type_name_of_expr env e in
+  let type_name, _an = type_name_of_expr env e in
   let unwrapped_e = unwrap_field_type loc name kind e in
   let default =
     match kind with
@@ -803,7 +880,10 @@ let inst_var_declaration
     Line (sprintf "%s: %s%s" var_name type_name default)
   ]
 
-let record env ~class_decorators loc name (fields : field list) an =
+let record env
+    ~class_decorators
+    ~(class_doc:node list)
+    loc name (fields : field list) an =
   let py_class_name = class_name env name in
   let trans_meth = env.translate_inst_variable () in
   let fields =
@@ -882,11 +962,36 @@ let record env ~class_decorators loc name (fields : field list) an =
       ]
     ]
   in
+  (* We create a docstring for the class corresponding to an ATD record from
+     - a generated title
+     - optional text coming from a <doc ...> annotation on the ATD definition
+     - optional text for each record field using ':param ...: ...' syntax
+       (Sphinx syntax)
+  *)
+  let record_doc =
+    trans_record_doc_to_docstring loc an
+  in
+  let field_doc =
+    List.map (trans_field_doc_to_docstring env trans_meth) fields
+    |> List.flatten
+  in
+  let docstring =
+    [
+      Line (sprintf {|"""Original type: %s = { ... }|} name);
+      Inline (blank_lines_before class_doc);
+      Inline class_doc;
+      Inline (blank_lines_before record_doc);
+      Inline record_doc;
+      Inline (blank_lines_before field_doc);
+      Inline field_doc;
+      Line {|"""|};
+    ]
+  in
   [
     Inline class_decorators;
     Line (sprintf "class %s:" py_class_name);
     Block (spaced [
-      Line (sprintf {|"""Original type: %s = { ... }"""|} name);
+      Inline docstring;
       Inline inst_var_declarations;
       Inline from_json;
       Inline to_json;
@@ -911,14 +1016,20 @@ class Foo:
   def from_json_string(x):
     ...
 *)
-let alias_wrapper env ~class_decorators name type_expr =
+let alias_wrapper env ~class_decorators ~class_doc name type_expr =
   let py_class_name = class_name env name in
-  let value_type = type_name_of_expr env type_expr in
+  let value_type, (loc, an) = type_name_of_expr env type_expr in
+  let type_doc = trans_alias_doc_to_docstring loc an in
   [
     Inline class_decorators;
     Line (sprintf "class %s:" py_class_name);
     Block [
-      Line (sprintf {|"""Original type: %s"""|} name);
+      Line (sprintf {|"""Original type: %s|} name);
+      Inline (blank_lines_before class_doc);
+      Inline class_doc;
+      Inline (blank_lines_before type_doc);
+      Inline type_doc;
+      Line {|"""|};
       Line "";
       Line (sprintf "value: %s" value_type);
       Line "";
@@ -951,15 +1062,19 @@ let alias_wrapper env ~class_decorators name type_expr =
 let case_class env ~class_decorators type_name
     (loc, orig_name, unique_name, an, opt_e) =
   let json_name = Atd.Json.get_json_cons orig_name an in
+  let case_doc = trans_case_doc_to_docstring loc an in
   match opt_e with
   | None ->
       [
         Inline class_decorators;
         Line (sprintf "class %s:" (trans env unique_name));
         Block [
-          Line (sprintf {|"""Original type: %s = [ ... | %s | ... ]"""|}
+          Line (sprintf {|"""Original type: %s = [ ... | %s | ... ]|}
                   type_name
                   orig_name);
+          Inline (blank_lines_before case_doc);
+          Inline case_doc;
+          Line {|"""|};
           Line "";
           Line "@property";
           Line "def kind(self) -> str:";
@@ -985,11 +1100,14 @@ let case_class env ~class_decorators type_name
         Inline class_decorators;
         Line (sprintf "class %s:" (trans env unique_name));
         Block [
-          Line (sprintf {|"""Original type: %s = [ ... | %s of ... | ... ]"""|}
+          Line (sprintf {|"""Original type: %s = [ ... | %s of ... | ... ]|}
                   type_name
                   orig_name);
+          Inline (blank_lines_before case_doc);
+          Inline case_doc;
+          Line {|"""|};
           Line "";
-          Line (sprintf "value: %s" (type_name_of_expr env e));
+          Line (sprintf "value: %s" (type_name_of_expr env e |> fst));
           Line "";
           Line "@property";
           Line "def kind(self) -> str:";
@@ -1057,7 +1175,7 @@ let read_cases1 env loc name cases1 =
             (class_name env name |> single_esc))
   ]
 
-let sum_container env ~class_decorators loc name cases =
+let sum_container env ~class_decorators ~class_doc loc name cases an =
   let py_class_name = class_name env name in
   let type_list =
     List.map (fun (loc, orig_name, unique_name, an, opt_e) ->
@@ -1091,11 +1209,19 @@ let sum_container env ~class_decorators loc name cases =
     else
       []
   in
+  let sum_doc =
+    trans_sum_doc_to_docstring loc an
+  in
   [
     Inline class_decorators;
     Line (sprintf "class %s:" py_class_name);
     Block [
-      Line (sprintf {|"""Original type: %s = [ ... ]"""|} name);
+      Line (sprintf {|"""Original type: %s = [ ... ]|} name);
+      Inline (blank_lines_before class_doc);
+      Inline class_doc;
+      Inline (blank_lines_before sum_doc);
+      Inline sum_doc;
+      Line {|"""|};
       Line "";
       Line (sprintf "value: Union[%s]" type_list);
       Line "";
@@ -1135,7 +1261,7 @@ let sum_container env ~class_decorators loc name cases =
     ]
   ]
 
-let sum env ~class_decorators loc name cases =
+let sum env ~class_decorators ~class_doc loc name cases an =
   let cases =
     List.map (fun (x : variant) ->
       match x with
@@ -1149,7 +1275,8 @@ let sum env ~class_decorators loc name cases =
     List.map (fun x -> Inline (case_class env ~class_decorators name x)) cases
     |> double_spaced
   in
-  let container_class = sum_container env ~class_decorators loc name cases in
+  let container_class =
+    sum_container env ~class_decorators ~class_doc loc name cases an in
   [
     Inline case_classes;
     Inline container_class;
@@ -1176,17 +1303,20 @@ let type_def env ((loc, (name, param, an), e) : A.type_def) : B.t =
     get_class_decorators an
     |> List.map (fun s -> Line ("@" ^ s))
   in
+  let class_doc =
+    trans_def_doc_to_docstring loc an
+  in
   let rec unwrap e =
     match e with
     | Sum (loc, cases, an) ->
-        sum env ~class_decorators loc name cases
+        sum env ~class_decorators ~class_doc loc name cases an
     | Record (loc, fields, an) ->
-        record env  ~class_decorators loc name fields an
+        record env ~class_decorators ~class_doc loc name fields an
     | Tuple _
     | List _
     | Option _
     | Nullable _
-    | Name _ -> alias_wrapper env ~class_decorators name e
+    | Name _ -> alias_wrapper env ~class_decorators ~class_doc name e
     | Shared _ -> not_implemented loc "cyclic references"
     | Wrap (loc, e, an) -> unwrap e
     | Tvar _ -> not_implemented loc "parametrized type"
@@ -1218,15 +1348,21 @@ let reserve_good_class_names env (items: A.module_body) =
     (fun (Type (loc, (name, param, an), e)) -> ignore (class_name env name))
     items
 
-let to_file ~atd_filename ~head (items : A.module_body) dst_path =
+let to_file ~atd_filename ~user_docstring ~head
+    (items : A.module_body) dst_path =
   let env = init_env () in
   reserve_good_class_names env items;
-  let head = List.map (fun s -> Line s) head in
   let python_defs =
     Atd.Util.tsort items
     |> List.map (fun x -> Inline (definition_group ~atd_filename env x))
   in
-  Line (fixed_size_preamble atd_filename) :: Inline head :: python_defs
+  [
+    Inline (preamble ~atd_filename ~user_docstring);
+    Inline head;
+    Line "";
+    Line "";
+    Inline (double_spaced python_defs);
+  ]
   |> double_spaced
   |> Indent.to_file ~indent:4 dst_path
 
@@ -1251,5 +1387,12 @@ let run_file src_path =
   in
   let full_module = Atd.Ast.use_only_specific_variants full_module in
   let (atd_head, atd_module) = full_module in
-  let head = Python_annot.get_python_json_text (snd atd_head) in
-  to_file ~atd_filename:src_name ~head atd_module dst_path
+  let user_docstring =
+    let loc, an = atd_head in
+    trans_head_doc_to_docstring loc an
+  in
+  let head =
+    Python_annot.get_python_json_text (snd atd_head)
+    |> List.map (fun s -> Line s)
+  in
+  to_file ~atd_filename:src_name ~user_docstring ~head atd_module dst_path
